#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 10 11:42:17 2021

@author: arnau/calum
"""

from fileinput import filename
from netCDF4 import Dataset
from os import listdir, walk
from os.path import isfile, join
from datetime import datetime, timedelta
from time import perf_counter
import numpy as np
import pandas as pd
import xarray as xr
import argparse
import sys
import os

import re

import dask.dataframe as dd


from geopandas import GeoDataFrame
from shapely.geometry import Polygon, Point


# import module from fortran
# path to the .so fortran
sys.path.insert(0, "../libs")
import elwriterdaf90

from grid_tools import *
from function_database import *


#variable we want to save
AODvarname = 'Aerosol_Optical_Thickness_550_Land_Ocean_Best_Estimate'
#variable Aerosol_Type_Land_Ocean values to filter. This variable filters only if = 0
filter1varname = 'Aerosol_Type_Land_Ocean'
#variable Aerosol_Optical_Thickness_QA_Flag_Ocean values to filter. This variable filters only if >=3 
filter2varname = 'Aerosol_Optical_Thickness_QA_Flag_Ocean'
#variable Aerosol_Optical_Thickness_QA_Flag_Land values to filter. This variable filters only if >=2 
filter3varname = 'Aerosol_Optical_Thickness_QA_Flag_Land'



def perform_rotation(tlm0d, tph0d, ds, aod_column, rotated):
    # Dropping a lot of nan values here
    df = ds[[aod_column, "Scan_Start_Time", "Longitude", "Latitude"]].to_dataframe().dropna(subset=[aod_column])
    geo_lon_in = df["Longitude"].values
    geo_lat_in = df["Latitude"].values
    invalid = 0
    
    if rotated:
        rotated_lat, rotated_lon = rotate_geo_to_rot(geo_lat_in,geo_lon_in, tlm0d , tph0d, westbound, southbound ,invalid, checkbounds = False)
    
        df["obslat"] = rotated_lat
        df["obslon"] = rotated_lon
    else:
        df["obslat"] = df["Longitude"]
        df["obslon"] = df["Latitude"]
    
   
    df["time"] = datetime.strptime("199301010000", "%Y%m%d%H%M") + df["Scan_Start_Time"]
    df = df.reset_index().drop(["Idx_Atrack", "Idx_Xtrack", "Scan_Start_Time"], axis=1)

    df = df.rename(columns={aod_column: "obs"})
    return df


def datemap(dates_slice, time_interval, filespath, lafrec = 'T'):
    #datetime mapping
    #group the obs in time slices. 
    # The name of the time slice (ex YYYYmmddHH) refers to the time YYYYmmddHH -1*lafrec_slice
    #the frequency interval implies that the time slice YYYYmmddHH-1*lafreq_slice do not contains obs for YYYYmmddHH
    #interval : [closed, open)
    """
    iDate (str) YYYYmmddHHMM / YYYY-mm-dd HH:mm (example)
    fDate (str) YYYYmmddHHMM / YYYY-mm-dd HH:mm (example)
    lafrec_slice (str)
    filespath (str)
    """
    
    # Find all files in our main dir and subdirs
    t0 = perf_counter()
    list_dir = []
    for (dirpath, dirnames, filenames) in walk(filespath):
        list_dir += [join(dirpath,file) for file in filenames]
    print("Time to list dir: {}".format(perf_counter() - t0))
    frequency = frequency_int(time_interval)

    # search coincidences with virtual files with frequency 1 min
    t0 = perf_counter()
    allfiles_slice = []
    for date in dates_slice:
        # Time interval centred around each hour
        left_border = (datetime.strptime(date, "%Y%m%d%H%M") - timedelta(hours=frequency/2)).strftime('%Y%m%d%H%M')
        right_border = (datetime.strptime(date, "%Y%m%d%H%M") + timedelta(hours=frequency/2)).strftime('%Y%m%d%H%M')

        dates_slice_virtual = pd.date_range(
            left_border,
            right_border,
            freq = lafrec, 
            closed = "right"
            ).strftime('%Y%j.%H%M')
        
        pattern = re.compile('|'.join(map(str, dates_slice_virtual)))
        date_files = list(filter(pattern.search, list_dir))
        allfiles_slice.extend(date_files)
        # Finds all files that are with this hour and adds them to 
        #allfiles_slice.extend([f for f in list_dir for date_virtual in dates_slice_virtual if date_virtual in f])
    print("Time for all files slice: {}".format(perf_counter() - t0))

    
    return allfiles_slice, list_dir

def df_calculations(df, left_border, right_border, center_lat_lon, aggregated_data, min_obs, unc_const):
    """
    Performing the aggregation and uncertainty calculations
    """
    #Prepare obs format, looking at particular time intervals
    df = df[df["time"].between(left_border, right_border, inclusive="right")]
    

    # If we want aggregated data
    if aggregated_data == True:
        df = grid.joining_observations_with_grid(df)

        # Produce a groupby of mean and count values
        count_df = df.groupby("grid_cell").count()
        mean_df = df.groupby("grid_cell").mean()
            
        # Reduce the df depending on min number of observations wanted in each cell
        count_df = count_df[count_df["obs"] >= min_obs]
        df = mean_df[mean_df.index.isin(count_df.index)]

        
        # Set the averaged lon/lat values to be the centre of the grid cell
        if center_lat_lon == True:
            df["obslat"] = df["lat_cen"]
            df["obslon"] = df["lon_cen"]



    #Need to update this to 
    # calculate the error sqrt((unca*AOD + uncb)**2 + uncc**2)
    df["obserr"] = np.sqrt((unc_const[0]*df["obs"]+unc_const[1])**2 + unc_const[2]**2)

    
    df = df.reset_index()
    df['obstype'] = int(30)
    df['obsn'] = int(1)
    df['obsid'] = int(34)
    df['obslev'] = int(-9999)
    

    return df


def getOptions(args=sys.argv[1:]):
    """
    Parsing arguments passed from the bash script
    """
    parser = argparse.ArgumentParser(description="Parses command.")
    parser.add_argument("-d", "--datadir", help="Where the original data is stored.")
    parser.add_argument("-f", "--finalobs", help="Where to save the final observations.")
    parser.add_argument("-a", "--unca", type=float, help="Uncertainty constant a.")
    parser.add_argument("-b", "--uncb", type=float, help="Uncertainty constant b.")
    parser.add_argument("-c", "--uncc", type=float, help="Uncertainty constant c.")
    parser.add_argument("-q", "--dateini", help="Initial date.")
    parser.add_argument("-w", "--datefin", help="Final date.")
    parser.add_argument("-e", "--lafreqslice", help="frequency of the final obs")
    parser.add_argument("-m", "--cenlon", type=float, help="center point longitudinal (E/W).")
    parser.add_argument("-l", "--cenlat", type=float, help="center point latitudinal (S/N).")
    parser.add_argument("-bs", "--southbound", type=float, help="South boundary")
    parser.add_argument("-bw", "--westbound", type=float, help="West boundary")
    parser.add_argument("-rlat", "--incrlat", type=float, help="Latitude resolution")
    parser.add_argument("-rlon", "--incrlon", type=float, help="Longitude resolution")
    parser.add_argument("-r", "--rotate", type=str2bool, nargs='?', const=True, default=False,
                        help="Rotate coordinates.")
    parser.add_argument("-avg", "--avgdata", type=str2bool, help="Do we want to average the observations over the grid cell")
    parser.add_argument("-min", "--minobs", type=int, help="Minimum number of observations per grid cell")
    parser.add_argument("-cen", "--cenlatlon", type=str2bool, help="Whether to centre the lat lon observations.")
    options = parser.parse_args(args)
    return options



if __name__ == "__main__":
    # Read in arguments -d ${datadir} -f ${finalobs} -a ${unca} -b ${uncb} -c ${uncc} -q ${dateini} -w ${datefin} -e ${lafreqslice} -m ${cenlon} -l ${cenlat} -r ${rotate}
    # * datadir - directory with the original files
    # * finalobs - output directory 
    # * unca, uncb, uncc -  Error : sqrt((unca*AOD + uncb)**2 + uncc)
    # * dateini - start date
    # * datefin - end date
    print("Parsing Arguments")
    options = getOptions(sys.argv[1:])

    # file = '/home/arnau/Escritorio/DADES-ARNAU/CURRO/BSC-CNS/VIIRS-AOD/dades/AERDB_L2_VIIRS_SNPP.A2020179.0936.001.2020179191236.nc'
    # filespath = '/media/esarchive/obs/nasa/viirs_snpp_aerdb_l2/original_files/VIIRS/2020/'
    filespath = options.datadir
    filesavepath = options.finalobs

    # This could have become obsolete
    #option to rotate coords
    # rot_coords = True
    rot_coords = options.rotate
    # tlm0d = 20.0 #center point longitudinal (E/W)
    tlm0d = options.cenlon
    # tph0d = 35.0 #center point latitudinal (S/N)
    tph0d = options.cenlat
    southbound = options.southbound
    westbound = options.westbound
    incrlat = options.incrlat
    incrlon = options.incrlon


    #uncertainty constants
    # unc_const = [0.2, 0.05, 0.02]
    unc_const = [options.unca, options.uncb, options.uncc]

    date_ini = options.dateini
    date_fin = options.datefin
    lafreq_slice = options.lafreqslice

    # Big endian =1, little endian !=1
    isbigendian = 1

    # Averaging over grid cells
    aggregated_data = options.avgdata

    # Min number of observation per grid cell
    min_obs = options.minobs

    # If we want to take the centre of the grid cells lat lon or averaged obs lat lon
    center_lat_lon = options.cenlatlon

    dates_slice = pd.date_range(date_ini, date_fin, freq=lafreq_slice, closed = None).strftime('%Y%m%d%H%M')

    print("Working out dates")
    files_sliced, list_dir = datemap(dates_slice, lafreq_slice, filespath)
    obsHeader = ['obstype', 'obslon', 'obslat', 'obsn', 'obsid', 'obslev', 'obs', 'obserr', 'time']

    
    t0 = perf_counter()
    # building monarch grid
    grid = Grid(tlm0d, tph0d, westbound, southbound, incrlat, incrlon)
    grid.creating_regular_coordinate_grid()
    grid.building_grid()
    print("Time for building the grid: {}".format(perf_counter() - t0))
    t0 = perf_counter()
    
    frequency = frequency_int(lafreq_slice)

    if not os.path.exists(filesavepath):
        print("Output directory doesn't exist, creating...")
        os.makedirs(filesavepath)
    # Produce the output files per date 
    for date in dates_slice:
        print(date)

        filewrite = filesavepath + '/obs' + str(date[:-2])

        # Look for files either side of the frequency with a samll overlap, because files have overlapping data
        # Time interval centred around each hour
        left_border = (datetime.strptime(date, "%Y%m%d%H%M") - timedelta(hours=frequency)).strftime('%Y%m%d%H%M')
        right_border = (datetime.strptime(date, "%Y%m%d%H%M") + timedelta(hours=frequency)).strftime('%Y%m%d%H%M')

        dates_slice_virtual = pd.date_range(
            left_border,
            right_border,
            freq = "T", 
            closed = "right"
            ).strftime('%Y%j.%H%M')
        
        pattern = re.compile('|'.join(map(str, dates_slice_virtual)))
        data_files = list(filter(pattern.search, list_dir))
        # If number of data files for these dates is 0 then create empty file and continue
        if len(data_files) == 0:
            os.system('touch {}'.format(filewrite))
            continue
        
        pandawrite = pd.DataFrame(columns = obsHeader)
        for file in data_files:
        
            # Reading every file needed
            ds = xr.open_dataset(file)
        
            ds = ds.where((ds[filter1varname] == 0) & ((ds[filter2varname]>= 3) | (ds[filter3varname] >= 2)))
        
            df = perform_rotation(tlm0d, tph0d, ds, AODvarname, rot_coords)
            pandawrite = pandawrite.append(df)

        pandawrite = pandawrite.reset_index(drop=True)
        # Cut down the dataframe here for all dates
        df = pandawrite

        frequency = frequency_int(lafreq_slice)
        left_border = (datetime.strptime(date, "%Y%m%d%H%M") - timedelta(hours=frequency/2)).strftime('%Y%m%d%H%M')
        right_border = (datetime.strptime(date, "%Y%m%d%H%M") + timedelta(hours=frequency/2)).strftime('%Y%m%d%H%M')
        print("Time periods")
        print(left_border)
        print(right_border)
        
        df = df_calculations(df, left_border, right_border, center_lat_lon, aggregated_data, min_obs, unc_const)

        print("Prepare Obs")
        t0 = perf_counter()
        obstype_in, obslon_in, obslat_in, obsn_in, obsid_in, obslev_in, obs_in, obserr_in, nentries, maxobs = prepare_obs(df)
        print("Time for prepare obs: {}".format(perf_counter() - t0))

        ## Output to netcdf
        dataframe_to_nc(df, filewrite)

        # Save obs in binary format
        print("Fortran write")
        t1 = perf_counter()
        # If no data for date, then create an empty file 
        
        if nentries == 0:
            os.system('touch {}'.format(filewrite))
        else:
            elwriterdaf90.writeobs4da_cams(filewrite, obstype_in, obslon_in, obslat_in, obsn_in,
                                        obsid_in, obslev_in, obs_in, obserr_in, 
                                        int(isbigendian), len(filewrite))
        print("Time for fortran write: {}".format(perf_counter() - t1))