#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Nov  2 11:58:46 2020

@author: arnau
"""
import subprocess
import argparse
import sys
import os
import json

import numpy as np
import pandas as pd

from datetime import timedelta
from numpy import f2py
from os import listdir, makedirs, walk
from os.path import isfile, join, exists, getsize
from tools.CAMS_standards import standard_parameters

#import module from fortran
# path to the .so fortran
path_to_so = './tools/'
sys.path.insert(0, path_to_so)
import elwriterdaf90
import rotated


#Function to introduce boolean arguments 
def str2bool(v):
    if isinstance(v, bool):
       return v
    if v.lower() in ('yes', 'true', 't', 'y', '1'):
        return True
    elif v.lower() in ('no', 'false', 'f', 'n', '0'):
        return False
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


def prepare_obs(observation):
    
    """
    group observation files by type, lon and lat and check if there are more than 1 
    value for each lon lat 
    
    obstype, obslon, obslat obsn will be 1d arrays
    obsid, obslev, obs and obserr will be matrixes with 0 value where we dont have 
    more than 1 obs
        observation = dataFrame with obsHeader
    """
    gb = observation.groupby(['obstype','obslon','obslat'])
    nentries = len(gb.groups)
    
    
    obstype_in = np.array([l[0] for l in gb.groups], dtype = 'int')
    obslon_in = np.array([l[1] for l in gb.groups], dtype = 'float64')
    obslat_in = np.array([l[2] for l in gb.groups], dtype = 'float64')
    obsn_in = np.array([int(len(gb.get_group(i)['obsn'])) for i in gb.groups], dtype = 'int')
    
    maxobs = max(obsn_in)
    
    #make an array[nentries, maxobs] and if we do not have obs put 0
    obsid_in =np.zeros([nentries, maxobs])
    obsid_list = [gb.get_group(i)['obsid'].values for i in gb.groups]
    obslev_in  = np.zeros([nentries,maxobs])
    obslev_list = [gb.get_group(i)['obslev'].values for i in gb.groups]
    obs_in = np.zeros([nentries, maxobs])
    obs_list = [np.array(gb.get_group(i)['obs'].values, dtype='float64') for i in gb.groups]
    obserr_in = np.zeros([nentries, maxobs])
    obserr_list = [np.array(gb.get_group(i)['obserr'].values, dtype = 'float64') for i in gb.groups]
    for row in range(nentries):
        auxid = np.array(obsid_list[row])
        auxid.resize(obsid_in[row].shape)
        obsid_in[row] = obsid_in[row] + auxid
        
        auxlev = np.array(obslev_list[row])
        auxlev.resize(obslev_in[row].shape)
        obslev_in[row] = obslev_in[row] + auxlev
        
        aux = np.array(obs_list[row])
        aux.resize(obs_in[row].shape)
        obs_in[row] = obs_in[row] + aux
        
        auxerr = np.array(obserr_list[row])
        auxerr.resize(obserr_in[row].shape)
        obserr_in[row] = obserr_in[row] + auxerr

    return obstype_in, obslon_in, obslat_in, obsn_in, obsid_in, obslev_in, obs_in, obserr_in, nentries, maxobs


def datemap(iDate, fDate, lafrec_slice, filespath, lafrec = 'D'):
    #datetime mapping
    #group the obs in time slices. 
    """
    iDate (str) YYYYmmddHHMM / YYYY-mm-dd HH:mm (example)
    fDate (str) YYYYmmddHHMM / YYYY-mm-dd HH:mm (example)
    lafrec_slice (str)
    filespath (str)
    """
    fDateFloor = fDate.floor('d')
    fDate = fDate + timedelta(days = 1) if fDate != fDateFloor else fDate
    
    dates_slice = pd.date_range(pd.to_datetime(iDate) - timedelta(days = 1), fDate, freq = lafrec_slice).strftime('%Y%m%d')
    
    # Find all files in our main dir and subdirs
    list_dir = []
    for (dirpath, dirnames, filenames) in walk(filespath):
        list_dir += [join(dirpath,file) for file in filenames]
    print(list_dir)
    
    # search coincidences with virtual files with frequency 1 day
    allfiles_slice = [f for f in list_dir for date in dates_slice if date in f]
    return allfiles_slice


def read_and_process_data(iDate, fDate, dataDir, elements, an_re, unit):
    element_analysis = {'o3':'O3', 'no2':'NO2', 'so2':'SO2', 'co':'CO', 'pm10':'PM10', 'pm2p5':'PM25'}
    allfiles = datemap(iDate, fDate, 'D', dataDir)
    print("Files to be downloaded")
    print(allfiles)
    for file in allfiles:
        # reading in file
        df = pd.read_csv(
            file, 
            delimiter=';', 
            dtype={
                'YEAR':str, 
                "MONTH":str, 
                'DAY':str, 
                'HOUR':str
                })
        
        # Add necessary data columns
        # Create a new row of elements that correspond to those being passed by the user
        df['element_con'] = df['PARAMETER'].replace(element_analysis)
        df = df.rename(columns={"CONCENTRATION(kg/m3)":"CONCENTRATION"})
        # date e.g. 2021, 05, 07, 00 to 2021050700
        df['date'] = df[['YEAR', 'MONTH', 'DAY', 'HOUR']].agg(''.join, axis=1)


        # Add all observational events for each specific date and element
        for e in elements:
            print(e)
            for date in dates:
                print(date)
                cut_df = df[(df.element_con == e) & (df.date == date)]
                print(cut_df)


                # Apply boundaries for the values
                cut_df = cut_df[cut_df['CONCENTRATION'].between(
                    standard_parameters[e]['extreme_lower_limit_{}'.format(an_re)],
                    standard_parameters[e]['extreme_upper_limit_{}'.format(an_re)],
                )]
                # Observational columns which will be outputted after refactoring
                cut_df['obslon']= cut_df['LON'].astype(float) #col 1
                cut_df['obslat'] = cut_df['LAT'].astype(float) # col 2
                # Change units

                #cut_df['obs'] = cut_df['CONCENTRATION(kg/m3)'].astype(float)*eval(unit[e_an][1]) #col 10 
                cut_df['obs'] = cut_df['CONCENTRATION'].astype(float)*eval(unit[e][1]) #col 10 
                cut_df['obserr'] = np.maximum(unc_by_element[e][1],cut_df['obs']*unc_by_element[e][0])
                print("Here")
                print(cut_df)
                global obsnorot
                obsnorot[e][date] = obsnorot[e][date].append(cut_df[obsnorotHeader])


def uncertainty_by_element(options):
    """
    Function for assigning uncertainties to specific elements instead of a generic constant.
    Joins together uncb and uncc dictionaries into single dictionary as described below.
    Throws an error if an element is in 'elements' but hasn got assigned uncertainty values.

    
    returns: dictionary in the form {element: (uncb for element, uncc for element)}
    """
    # get a dictionary of argparse arguments and choose variables needed for uncertanties
    var_dict = vars(options)
    elements = var_dict["element"]

    uncb = var_dict["uncb"]
    uncc = var_dict["uncc"]
    unc_by_element = {}
    for e in elements:
        try:
            unc_by_element[e] = (uncb[e], uncc[e])

        except KeyError:
            print("Element {} has been passed without an uncertainty. Uncertainties passed uncb={}, uncc={}".format(e, uncb, uncc))
            raise
    return unc_by_element


def getOptions(args=sys.argv[1:]):
    parser = argparse.ArgumentParser(description="Parses command.")
    parser.add_argument("-d", "--datadir", help="Where the original data is stored.")
    parser.add_argument("-s", "--stationsdir", help="Where the info about the stations is stored.")
    parser.add_argument("-u", "--unitchange", help="Where the file with the conversion factors is stored.")
    parser.add_argument("-e", "--element", nargs = '+', help="Which element to process.")
    #if fortran script is not compiled uncomment next line and correspondent block code
    # parser.add_argument("-c", "--compile", help="Where the code to compile is.")
    parser.add_argument("-x", "--execute", help="Where the fortran code (rotate) to execute is.")
    parser.add_argument("-o", "--obsnorotated", help="Where to save the no rotated observations.")
    parser.add_argument("-f", "--obsrotated", help="Where to save the rotated observations in binary format.")
    parser.add_argument("-a", "--saveAllinOne", type=str2bool, nargs='?', const=True, default=False,
                        help="Save final obs separated by elements or all in one file.")
    parser.add_argument("-m", "--cenlon", help="center point longitudinal (E/W).")
    parser.add_argument("-l", "--cenlat", help="center point latitudinal (S/N).")
    parser.add_argument("-w", "--wbound", help="Western boundary.")
    parser.add_argument("-q", "--sbound", help="Southern boundary.")
    parser.add_argument("-b", "--uncb", type=json.loads, help="Uncertainty constant b.")
    parser.add_argument("-c", "--uncc", type=json.loads, help="Uncertainty constant c.")
    parser.add_argument("-i", "--startdate", help="Initial date in YYYY-MM-DD format.")
    parser.add_argument("-t", "--enddate", help="End date in YYYY-MM-DD format.")
    parser.add_argument("-r", "--anre", help="If the obs are for analysis or reanalysis")
    options = parser.parse_args(args)
    return options

if __name__ == "__main__":
    print('STARTING')
    options = getOptions(sys.argv[1:])

    print('PARSING ARGUMENTS')
    # Creating variables from arguments passed from automonarch
    dataDir = options.datadir
    unitChangeFile = options.unitchange
    elements = options.element
    obs_no_rotated_save = options.obsnorotated
    obs_final_save = options.obsrotated
    fortranexec = options.execute
    saveallinone = options.saveAllinOne
    tlm0d = options.cenlon
    tph0d = options.cenlat
    wbd = options.wbound
    sbd = options.sbound
    an_re = options.anre
    iDate = options.startdate
    fDate = options.enddate

    # Creating output directories if they don't already exist
    if not exists(obs_no_rotated_save):
        makedirs(obs_no_rotated_save)

    if not exists(obs_final_save):
        makedirs(obs_final_save)

    #uncertainty constants
    unc_by_element = uncertainty_by_element(options)

    #IDs in the DA code
    id_data = {'in_situ':90, 'O3':91, 'NO2':92, 'SO2':93, 'CO':94, 'PM10':95, 'PM25':96}
    aidi = pd.Series(id_data)

    #read unit change file
    unit = pd.read_csv(unitChangeFile, sep = ' ', index_col = 0, header = None).T
    print(unit)

    if an_re != 'an' and an_re != 're':
        raise Exception('anre variable must be an or re, not {}'.format(an_re))

    #Make sure dates are in correct format
    iDate = pd.to_datetime(iDate)
    fDate = pd.to_datetime(fDate)
    lafrec = 'H'
    dates = pd.date_range(iDate, fDate, freq = lafrec).strftime('%Y%m%d%H')
    dates_years = pd.date_range(iDate, fDate, freq = 'H').strftime('%Y').drop_duplicates()

    # For each element and date create empty dataframe to be filled later
    obs = dict()
    obsnorot = dict()
    obsHeader = ['obstype', 'obslon', 'obslat', 'obsn', 'obsid', 'obslev', 'obs', 'obserr']
    obsnorotHeader = ['date', 'obslon', 'obslat', 'obs', 'obserr']
    for e in elements:
        obsnorot[str(e)]=dict()
        for d in dates:
            # obs[str(d)] = pd.DataFrame(columns = obsHeader)
            obsnorot[str(e)][str(d)] = pd.DataFrame(columns = obsnorotHeader)
    print('READING DATA FILES')
    #read the data files

    ### Read in dataand preprocess it adding to obs dictionary
    read_and_process_data(iDate, fDate, dataDir, elements, an_re, unit)

    print('SAVING NO ROTATED OBSERVATIONS')                   
    #save 
    for e in elements:
        obsDirElement = join(obs_no_rotated_save, str(e))
        if not exists(obsDirElement):
            makedirs(obsDirElement)
        for key in obsnorot[str(e)]:
            obsfile = open(obsDirElement + '/' + str(key) + '_natural_glo.txt','w')
            obsfile.write(str(len(obsnorot[str(e)][key]))+'\n')
            obsnorot[str(e)][key].to_csv(obsfile, sep = ',', index=False, header=False)
            obsfile.close()


    #Run fortran script (ROTATE)
    print('ROTATING')
    for e in elements:
        obsDirElement = join(obs_no_rotated_save, str(e))
        for key in obsnorot[str(e)]:
            entrada = obsDirElement + '/' + str(key) + '_natural_glo.txt'
            sortida1 = obsDirElement + '/' + str(key) + '_rotated_reg.txt'
            sortida2 = obsDirElement + '/' + str(key) + '_natural_reg.txt'
            subprocess.run([fortranexec, entrada, sortida1, sortida2, tlm0d, tph0d, wbd, sbd])

    #write da binary format with fortran
    print('WRITING IN BINARY FORMAT')

    isbigendian = 1

    #if all observations in the same file, initialize the dict with dates as keys
    if saveallinone:
        pandawrite = dict()
        for e in elements:
            for key in obsnorot[str(e)]:
                pandawrite[str(key)] = pd.DataFrame(columns = obsHeader)
            
    #if the user wants to save all obs in the same file 
    if saveallinone:
        obsfinalDirElement = obs_final_save
    
        for e in elements:
            obsDirElement = join(obs_no_rotated_save, str(e))
        
            if not exists(obsfinalDirElement):
                makedirs(obsfinalDirElement)
    
            for key in obsnorot[str(e)]:            
                fileread = obsDirElement + '/' + str(key) + '_rotated_reg.txt'
            
                pandaread = pd.read_csv(fileread, sep = '\s+', header = None, engine='python', dtype=str)
                datawrite = {'obstype':90, 'obslon':pandaread[1], 
                        'obslat':pandaread[2], 'obsn':1, 
                        'obsid': aidi[e], 'obslev': 1, 'obs':pandaread[3],
                        'obserr':pandaread[4]}
            
                pandawrite[str(key)] = pandawrite[str(key)].append(pd.DataFrame(datawrite), ignore_index = True)
        # Write Binary
        for date in pandawrite:
            filewrite = obsfinalDirElement + '/obs' + str(date)
            #prepare observations
            obstype_in, obslon_in, obslat_in, obsn_in, obsid_in, obslev_in, obs_in, obserr_in, nentries, maxobs = prepare_obs(pandawrite[str(date)])
            #write observations
            elwriterdaf90.writeobs4da_cams(filewrite, obstype_in, obslon_in, obslat_in, obsn_in,
                                       obsid_in, obslev_in, obs_in, obserr_in, 
                                       int(isbigendian), len(filewrite))
        
    else:
        for e in elements:       
            obsfinalDirElement = join(obs_final_save, str(e))
            obsDirElement = join(obs_no_rotated_save, str(e))
            if not exists(obsfinalDirElement):
                makedirs(obsfinalDirElement)          
            for key in obsnorot[str(e)]:
                fileread = obsDirElement + '/' + str(key) + '_rotated_reg.txt'
                if getsize(fileread) != 0:
                    filewrite = obsfinalDirElement + '/obs' + str(key)

                    pandaread = pd.read_csv(fileread, sep = '\s+', header = None, engine='python', dtype=str)
                    datawrite = {'obstype':90, 'obslon':pandaread[1],
                          'obslat':pandaread[2], 'obsn':1,
                          'obsid': aidi[e], 'obslev': 1, 'obs':pandaread[3],
                          'obserr':pandaread[4]}
                    pandawrite = pd.DataFrame(datawrite)
                    # Write binary format using fortran function
                    elwriterdaf90.writeobs4da_cams(filewrite,pandawrite['obstype'].values.astype('int'),
                                       pandawrite['obslon'].values.astype('float64'),
                                       pandawrite['obslat'].values.astype('float64'),
                                       pandawrite['obsn'].values.astype('int'),
                                       pandawrite['obsid'].values.astype('int'),
                                       pandawrite['obslev'].values.astype('int'),
                                       pandawrite['obs'].values.astype('float64'),
                                       pandawrite['obserr'].values.astype('float64'),
                                       int(isbigendian), len(filewrite))
                    os.system('chmod 775 -R {}'.format(filewrite))


