#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
@author: calum
"""

import geopandas as gpd
import numpy as np

from geopandas import GeoDataFrame
from shapely.geometry import Polygon, Point


class Grid:
    """
    An Object to create the grid
    
    The grid is automatically rotated if at a regional scale
    and kept as regular at the global.
    """

    def __init__(self, cenlon, cenlat, westbound, southbound, incrlat, incrlon):
        self.attributes = {'new_pole_longitude_degrees': -180 + cenlon, 'new_pole_latitude_degrees': cenlat,
                      'centre_lat': cenlat, 'centre_lon': cenlon, 'west_boundary': westbound,
                      'south_boundary': southbound, 'inc_rlat': incrlat, 'inc_rlon': incrlon,
                      'n_lat': int((abs(southbound) / incrlat) * 2 + 1), # Need to imporve this, so that isn't just half
                      'n_lon': int((abs(westbound) / incrlon) * 2 + 1), 'crs': {'init': 'epsg:4326'}}


    def creating_regular_coordinate_grid(self):
        """
        Center latitude and Longitudes are created from the southbound and westbound attributes.
        With the (southbound, westbound) being the center of the most south westerly grid cell

        It is then converted into a numpy array (4, n) of all the corner points of all grid cells
        """
        
        self.center_latitudes = np.linspace(self.attributes['south_boundary'], self.attributes['south_boundary'] +
                                       (self.attributes['inc_rlat'] * (self.attributes['n_lat'] - 1)),
                                       self.attributes['n_lat'], dtype=float) 
        self.center_longitudes = np.linspace(self.attributes['west_boundary'], self.attributes['west_boundary'] +
                                        (self.attributes['inc_rlon'] * (self.attributes['n_lon'] - 1)),
                                        self.attributes['n_lon'], dtype=float)
        
        
        counter = 0
        self.coords = np.empty((self.center_latitudes.shape[0] * self.center_longitudes.shape[0], 2))
        for i in range(self.center_longitudes.shape[0]):
            for j in range(self.center_latitudes.shape[0]):
                self.coords[counter, 0] = self.center_longitudes[i]
                self.coords[counter, 1] = self.center_latitudes[j]
                counter += 1

        self.b_lats = self.create_bounds(self.coords[:, 1], self.attributes['inc_rlat'], number_vertices=4, inverse=True)
        self.b_lons = self.create_bounds(self.coords[:, 0], self.attributes['inc_rlon'], number_vertices=4)
        return True


    @staticmethod
    def create_bounds(coordinates, inc, number_vertices=2, inverse=False):
        """
        Calculate the vertices coordinates.

        :param coordinates: Coordinates in degrees (latitude or longitude)
        :type coordinates: numpy.array

        :param inc: Increment between center values.
        :type inc: float

        :param number_vertices: Non mandatory parameter that informs the number of vertices that must have the
                boundaries (by default 2).
        :type number_vertices: int

        :param inverse: For some grid latitudes.
        :type inverse: bool

        :return: Array with as many elements as vertices for each value of coords.
        :rtype: numpy.array
        """
        # Create new arrays moving the centers half increment less and more.
        coords_left = coordinates - inc / 2
        coords_right = coordinates + inc / 2

        # Defining the number of corners needed. 2 to regular grids and 4 for irregular ones.
        if number_vertices == 2:
            # Create an array of N arrays of 2 elements to store the floor and the ceil values for each cell
            bound_coords = np.dstack((coords_left, coords_right))
            bound_coords = bound_coords.reshape((len(coordinates), number_vertices))
        elif number_vertices == 4:
            # Create an array of N arrays of 4 elements to store the corner values for each cell
            # It can be stored in clockwise starting form the left-top element, or in inverse mode.
            if inverse:
                bound_coords = np.dstack((coords_left, coords_left, coords_right, coords_right))
            else:
                bound_coords = np.dstack((coords_left, coords_right, coords_right, coords_left))
        else:
            print('The number of vertices of the boundaries must be 2 or 4.')
        return bound_coords


    def building_grid(self):
        """
        Here we build the grid using Shapely polygons. And attribute it to a geopandas dataframe

        This is also where we decide whether to use a regular grid for global or a rotated grid for regional

        """


        # May need to add centre lats and centre lons to this dataframe as well

        # Re-shape
        aux_b_lats = self.b_lats.squeeze()
        aux_b_lons = self.b_lons.squeeze()

        geometry = []
        # Create one dataframe with 8 columns, 4 points with two coordinates each one
        for i in range(aux_b_lons.shape[0]):
            geometry.append(Polygon([(aux_b_lons[i, 0], aux_b_lats[i, 0]),
                                     (aux_b_lons[i, 1], aux_b_lats[i, 1]),
                                     (aux_b_lons[i, 2], aux_b_lats[i, 2]),
                                     (aux_b_lons[i, 3], aux_b_lats[i, 3]),
                                     (aux_b_lons[i, 0], aux_b_lats[i, 0])]))

        self.gdf = GeoDataFrame(index=range(aux_b_lons.shape[0]), geometry=geometry, crs='epsg:4326')
        # Append the center lat and lon values to this dataframe
        self.gdf["lon_cen"] = self.coords[:, 0]
        self.gdf["lat_cen"] = self.coords[:, 1]
        
        return True

    def joining_observations_with_grid(self, df):
        """
        Here we join the observations with the grid created.
        """

        # Creating shapely point coordinates out of observation lat and lon values
        df['coords'] = list(zip(df['obslon'],df['obslat']))
        df['coords'] = df['coords'].apply(Point)

        points = gpd.GeoDataFrame(df, geometry='coords', crs='epsg:4326')

        self.gdf["grid_cell"] = ["grid_cell_{}".format(i+1) for i in range(len(self.gdf))]

        result = gpd.tools.sjoin(points, self.gdf, how="left")
        # Need to remove NaNs if observations are not in the grid
        result = result.dropna(subset=["grid_cell"])
        
        return result
    



####### Rotation functions, copied from Jeronimo's scripts
###### je@bsc-es. translated form Monarch-NMMB and preproc fortran codes from Z. Janjic.  

def rotate_geo_to_rot(geo_lat_in,geo_lon_in, tlm0d , tph0d, wbd, sbd ,invalid, checkbounds = False):
    """
    Get the input ready to go through the mathematical formulation of the rotation

    Rotation: input is cartesian lon and lat values
    """

    geo_lat =np.atleast_2d(geo_lat_in).astype('float')
    geo_lon = np.atleast_2d(geo_lon_in).astype('float')
  
    dtr=np.pi/180.
    rtd=180./np.pi
    xlon = geo_lon
    xlat = geo_lat
  
    tph0=tph0d*dtr #!rad of the centre lat
    ctph0=np.cos(tph0) #!cos of the centre lat
    stph0=np.sin(tph0) #!sin of the centre lat
    tlm, tph  = tll(xlon,xlat,tlm0d,dtr,ctph0,stph0)
    rlon=tlm*rtd
    rlat=tph*rtd

    if checkbounds:
        # Checks the boundaries for any invalid lat and lon values
        # Before setting them to be the invalid variable
        bcw=np.abs(wbd)
        bcs=np.abs(sbd)
        
        b_0 = np.logical_and(  np.abs(rlon) < np.abs(bcw) , np.abs(rlat)<np.abs(bcs)  )
        
        rlat[~b_0] = invalid
        rlon[~b_0] = invalid

    rlat = rlat.squeeze()
    rlon = rlon.squeeze()
    
    return rlat,rlon



def tll(xlon,xlat,tlm0d,dtr,ctph0,stph0):
    """
        # from f90 version: 
        ****************************************************************
        *                                                              *
        *  programer: z. janjic, shmz, feb. 1981                       *
        *  ammended:  z. janjic, ncep, jan. 1996                       *
        *                                                              *
        *  transformation from lat-lon to rotated lat-lon coordinates  *
        ****************************************************************
        ****************************************************************
        *  tlm   - transformed longitude, rad.                         *
        *  tph   - transformed latitude, rad.                          *
        *  tlm0d - the angle of rotation of the transformed lat-lon    *
        *          system in the longitudinal direction, degs          *
        *  ctph0 - cos(tph0), tph0 is the angle of rotation of the     *
        *          transformed lat-lon systemn in the latitudinal      *
        *          direction, precomputed                              *
        *  stph0 - sin(tph0), tph0 is the angle of rotation of the     *
        *          transformed lat-lon systemn in the latitudinal      *
        *          direction, precomputed                              *
        *  almd  - geographical longitude, degs, range -180.,180       *
        *  aphd  - geographical latitude,  degs, range - 90., 90.,     *
        *          poles are singular                                  *
        ****************************************************************
    """
    relm=(xlon-tlm0d)*dtr #!distance from the centre lon (in rad)

    srlm=np.sin(relm) #!sin of this distance
    crlm=np.cos(relm) #!cos of this distance

    aph=xlat*dtr #!lat in rad 

    sph=np.sin(aph) #!sin of lat
    cph=np.cos(aph) #!cos of lat

    cc=cph*crlm #!cos of lat times cos of lon distance
    anum=cph*srlm #!cos of lat times sin of lon distance
    denom=ctph0*cc+stph0*sph #!cos of the centre lat times cc plus sin of the centre lat times sin of lat

    tlm=np.arctan2(anum,denom)
    tph=np.arcsin(ctph0*sph-stph0*cc)

    return tlm, tph



def rotate_rot_to_geo(rlat_in,rlon_in, tlm0d, tph0d):
    """
    Get the input ready to go through the mathematical formulation of the rotation

    Inverse Rotation: input is rotated lon and lat values
    """
    rlat=np.atleast_2d(rlat_in)
    rlon=np.atleast_2d(rlon_in)
  
    dtr=np.pi/180.
  
    tlm = rlon*dtr
    tph = rlat*dtr
    tph0=tph0d*dtr #!rad of the centre lat

    ctph0=np.cos(tph0) #!cos of the centre lat
    stph0=np.sin(tph0) #!sin of the centre lat

    almd , aphd = rtll(tlm,tph,tlm0d,dtr, ctph0,stph0)
    return aphd, almd


def rtll(tlm_in,tph_in,tlm0d, dtr, ctph0,stph0) :
    """
        subroutine rtll(tlm,tph,tlm0d,ctph0,stph0,almd,aphd)
        ****************************************************************
        *                                                              *
        *  programer: z. janjic, shmz, feb. 1981                       *
        *  ammended:  z. janjic, ncep, jan. 1996                       *
        *                                                              *
        *  transformation from rotated lat-lon to lat-lon coordinates  *
        ****************************************************************
        ****************************************************************
        *  tlm   - transformed longitude, rad.                         *
        *  tph   - transformed latitude, rad.                          *
        *  tlm0d - the angle of rotation of the transformed lat-lon    *
        *          system in the longitudinal direction, degs          *
        *  ctph0 - cos(tph0), tph0 is the angle of rotation of the     *
        *          transformed lat-lon systemn in the latitudinal      *
        *          direction, precomputed                              *
        *  stph0 - sin(tph0), tph0 is the angle of rotation of the     *
        *          transformed lat-lon systemn in the latitudinal      *
        *          direction, precomputed                              *
        *  almd  - geographical longitude, degs, range -180.,180       *
        *  aphd  - geographical latitude,  degs, range - 90., 90.,     *
        *          poles are singular                                  *
        ****************************************************************
    """
    tlm = np.atleast_2d(tlm_in)
    tph = np.atleast_2d(tph_in)

    stlm=np.sin(tlm)
    ctlm=np.cos(tlm)
    stph=np.sin(tph)
    ctph=np.cos(tph)

    sph=ctph0*stph+stph0*ctph*ctlm
    sph=np.minimum(sph,1.)
    sph=np.maximum(sph,-1.)
  
    aph=np.arcsin(sph)
    aphd=aph/dtr
    anum=ctph*stlm
    denom=(ctlm*ctph-stph0*sph)/ctph0

    relm=np.arctan2(anum,denom)
    almd=relm/dtr+tlm0d

    almd[almd >  180.] = almd[almd >  180.]-360.
    almd[almd < -180.] = almd[almd < -180.]+360.


    return almd.squeeze(), aphd.squeeze()



